import time
import os
from datetime import datetime, timedelta
import boto3
import click
import pytz
from brume.color import Color

from botocore.exceptions import ClientError

client = boto3.client('cloudformation')
TZ = pytz.timezone('UTC')


def env(key):
    return os.getenv(key, None)


def make_tags(tags_list):
    return [{"Key": k, "Value": v} for k, v in tags_list.items()]


def make_parameters(tags_list):
    return [{"ParameterKey": k, "ParameterValue": v} for k, v in tags_list.items()]


def stack_walker(outputs, stack, collector):
    try:
        description = client.describe_stacks(StackName=stack)['Stacks'][0]
        stackName = description['StackName']
        collector(outputs, description)
        substacks = [s for s in client.describe_stack_resources(StackName=stack)['StackResources'] if s['ResourceType'] == 'AWS::CloudFormation::Stack']
        for s in substacks:
            outputs[s['LogicalResourceId']] = {}
            stack_walker(outputs[s['LogicalResourceId']], s['PhysicalResourceId'], collector)
        return outputs
    except ClientError as e:
        if 'does not exist' in e.message:
            click.secho('Stack [{}] does not exist'.format(stack), err=True, fg='red')
            exit(1)
        else:
            raise e

def output_collector(outputs, description):
    for o in description.get('Outputs', []):
        outputs[o['OutputKey']] = o['OutputValue']

def param_collector(outputs, description):
    for o in description.get('Parameters', []):
        outputs[o['ParameterKey']] = o['ParameterValue']


def stack_outputs(stack_name):
    """Return stack outputs."""
    return stack_walker({}, stack_name, output_collector)

class Stack():
    stack_name = None
    parameters = {}
    capabilities = []
    tags = {}
    on_failure = 'ROLLBACK'

    def __init__(self, conf, current_path):
        self.stack_name = str(conf['stack_name'])
        self.template_body = os.path.join(current_path, conf['template_body'])
        self.capabilities = conf.get('capabilities', self.capabilities)
        self.parameters = make_parameters(conf.get('parameters', self.parameters))
        self.tags = make_tags(conf['tags'])
        self.on_failure = conf.get('on_failure', self.on_failure)

        # Check the events 2 minutes before if the stack update starts way too soon
        self.update_started_at = datetime.now(TZ) - timedelta(minutes=2)

        self.stack_configuration = dict(
            StackName=self.stack_name,
            TemplateBody=open(self.template_body, 'r').read(),
            Parameters=self.parameters,
            Capabilities=self.capabilities,
            Tags=self.tags)

    def outputs(self):
        return stack_outputs(self.stack_name)

    def params(self):
        return stack_walker({}, self.stack_name, param_collector)

    @staticmethod
    def exists(stack_name):
        try:
            client.describe_stacks(StackName=stack_name)
        except ClientError as e:
            if 'AlreadyExistsException' in e.message:
                click.secho('Stack [{}] does not exist'.format(stack_name), err=True, fg='red')
                exit(1)
        else:
            return True

    def create(self):
        click.echo('Creating stack {}'.format(self.stack_name))
        try:
            client.create_stack(**self.stack_configuration)
            self.tail()
        except ClientError as e:
            if 'AlreadyExistsException' in e.message:
                click.secho('Stack [{}] already exists'.format(self.stack_name), err=True, fg='red')
                exit(1)

    def update(self):
        click.echo('Updating stack {}'.format(self.stack_name))
        try:
            response = client.update_stack(**self.stack_configuration)
            self.tail()
        except ClientError as e:
            if 'does not exist' in e.message:
                click.secho('Stack [{}] does not exist'.format(self.stack_name), err=True, fg='red')
                exit(1)
            if 'No updates are to be performed.' in e.message:
                click.secho('No updates are to be performed on stack [{}]'.format(self.stack_name), err=True, fg='red')
                exit(1)
            click.secho('Error {}'.format(e.message), err=True, fg='red')
            exit(1)
        except Exception as e:
             click.secho('Error {}'.format(e.message), err=True, fg='red')
             exit(1)

    def create_or_update(self):
        click.echo('Deploying stack {}'.format(self.stack_name))
        try:
            client.create_stack(**self.stack_configuration)
            self.tail()
        except ClientError as e:
            if 'does not exist' in e.message:
                click.secho('Stack [{}] does not exist'.format(self.stack_name), err=True, fg='red')
                exit(1)
            if 'No updates are to be performed.' in e.message:
                click.secho('No updates are to be performed on stack [{}]'.format(self.stack_name), err=True, fg='red')
                exit(1)
            click.secho(e.message, err=True, fg='red')
            click.secho('Stack {} already exists'.format(self.stack_name), err=True, fg='red')
            self.update()

    def delete(self):
        click.echo('Deleting stack {}'.format(self.stack_name))
        try:
            client.delete_stack(StackName=self.stack_name)
            self.tail()
        except ClientError:
            # if 'does not exist' in e.message:
            exit(1)

    def status(self):
        try:
            stacks = client.describe_stacks(StackName=self.stack_name)
            click.echo(Color.for_status(next(s['StackStatus'] for s in stacks['Stacks'])))
        except KeyError as e:
            click.secho(e, err=True, fg='red')
            exit(1)
        except ClientError as e:
            if 'does not exist' in e.message:
                click.secho('Stack [{}] does not exist'.format(self.stack_name), err=True, fg='red')
                exit(1)

    def get_events(self):
        events = client.describe_stack_events(StackName=self.stack_name)
        return reversed(events['StackEvents'])

    def tail(self, sleep_time=3):
        error = False
        seen = set()
        click.echo('Polling for events...')
        self.print_log_headers()
        events = self.get_events()
        while True:
            for e in events:
                if e['Timestamp'] < self.update_started_at:
                    seen.add(e['EventId'])
                if e['EventId'] in seen:
                    continue
                if 'FAILED' in e['ResourceStatus']:
                    error = True
                self._log_event(e)
                seen.add(e['EventId'])
            if self.stack_complete(e):
                if error:
                    exit(1)
                break
            time.sleep(sleep_time)
            events = self.get_events()

    def stack_complete(self, e):
        if e['LogicalResourceId'] == self.stack_name and e['ResourceStatus'].endswith('COMPLETE'):
            return True
        else:
            return False

    def print_log_headers(self):
        click.echo('{:23s} {:36s} {:30s} {:30s} {}'.format(
            'Timestamp', 'Status', 'Resource', 'Type', 'Reason'
        ))

    def _log_event(self, e):
        click.echo('{:23s} {:36s} {:30s} {:30s} {}'.format(
            e['Timestamp'].strftime('%Y-%m-%d %H:%M:%S UTC'),
            Color.for_status(e['ResourceStatus']),
            e['LogicalResourceId'],
            e['ResourceType'],
            e.get('ResourceStatusReason', ''),
        ))



import os
from unittest import TestCase, main
from brume.config import Config


class TestLoadConfig(TestCase):
    def test_load_config(self):
        current_path = os.path.dirname(os.path.abspath(__file__))
        config_template = os.path.join(current_path, 'test_load_config.yml')
        conf = Config.load(config_template)

        assert conf['region'] == 'eu-west-1'
        assert isinstance(conf['stack'], dict)
        assert isinstance(conf['templates'], dict)


if __name__ == '__main__':
    main()



import click

from glob import glob
from os import path
from yaml import dump

from config import Config
from stack import Stack
from template import Template
from assets import send_assets

# Brume Configuration
conf = {}
templates_config = {}
cf_config = {}
current_path = None

def process_assets():
    """Send assets if defined."""
    global conf
    global current_path
    if ('assets' in conf):
        assetsConfig = conf['assets']
        local_path = assetsConfig['local_path']
        s3_bucket = assetsConfig['s3_bucket']
        s3_path = assetsConfig['s3_path']
        click.echo("Processing assets from {} to s3://{}/{}".format(path.join(current_path, local_path), s3_bucket, s3_path))
        send_assets(local_path, s3_bucket , s3_path)

def collect_templates():
    """Convert every .cform template into a Template."""
    global templates_config
    global current_path
    templates = glob(path.join(current_path, templates_config.get('local_path', ''), '*.cform'))
    return [Template(t, templates_config, current_path) for t in templates]


def validate_and_upload():
    """Validate templates and upload new stack configuration."""
    templates = collect_templates()
    map(lambda t: t.validate(), templates)
    map(lambda t: t.upload(), templates)
    process_assets()

def newStack():
    """Instanciante main Stack."""
    global cf_config
    global current_path
    return Stack(cf_config, current_path)

@click.command()
def config():
    """Print the current stack confguration."""
    global conf
    print(dump(conf))


@click.command()
def create():
    """Create a new CloudFormation stack."""
    validate_and_upload()
    newStack().create()


@click.command()
def update():
    """Update an existing CloudFormation stack."""
    validate_and_upload()
    newStack().update()


@click.command()
def deploy():
    """Create or update a CloudFormation stack."""
    validate_and_upload()
    newStack().create_or_update()


@click.command()
def delete():
    """Delete a CloudFormation stack."""
    newStack().delete()


@click.command()
def status():
    """Get the status of a CloudFormation stack."""
    newStack().status()


@click.command()
def outputs():
    """Get the full list of outputs of a CloudFormation stack."""
    outputs = newStack().outputs()
    print dump(outputs, default_flow_style=False)


@click.command()
def parameters():
    """Get the full list of parameters of a CloudFormation stack."""
    global cf_config
    parameters = newStack().params()
    print dump(parameters, default_flow_style=False)


@click.command()
def validate():
    """Validate CloudFormation templates."""
    templates = collect_templates()
    return map(lambda t: t.validate(), templates)


@click.command()
def upload():
    """Upload CloudFormation templates and assets to S3."""
    process_assets()
    templates = collect_templates()
    return map(lambda t: t.upload(), templates)


@click.group()
@click.option('--configuration', default='brume.yml')
def cli(configuration):
    global conf
    global templates_config
    global cf_config
    global current_path
    conf = Config.load(configuration)
    templates_config = conf['templates']
    cf_config = conf['stack']
    current_path = path.dirname(configuration)
    pass


cli.add_command(create)
cli.add_command(update)
cli.add_command(deploy)
cli.add_command(upload)
cli.add_command(delete)
cli.add_command(validate)
cli.add_command(config)
cli.add_command(status)
cli.add_command(outputs)
cli.add_command(parameters)

if __name__ == '__main__':
    cli()
